#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Test for functions to save/load Pyomo model state to a dict or json
"""

import unittest
import os
from tempfile import mkdtemp

from pyomo.environ import *
from idaes.core.util import to_json, from_json, StoreSpec
from idaes.core.util.model_serializer import _only_fixed
import shutil
import pytest

__author__ = "John Eslick"


class TestModelSerialize(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.dirname = mkdtemp()
        cls.fname = os.path.join(cls.dirname, "crAzYStuff1010202030.json")

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.dirname)

    def tearDown(self):
        try:
            os.remove(self.fname)
        except:
            pass

    def setup_model01(self, name=None):
        if name is not None:
            model = ConcreteModel(name)
        else:
            model = ConcreteModel()
        model.b = Block([1, 2, 3])
        a = model.b[1].a = Var(bounds=(-100, 100), initialize=2)
        b = model.b[1].b = Var(bounds=(-100, 100), initialize=20)
        x = model.x = BooleanVar(initialize=True)
        model.b[1].c = Constraint(expr=b == 10 * a)
        a.fix(2)
        return model

    def setup_model01b(self):
        model = ConcreteModel()
        model.b = Block(["1", "2", "3"])
        a = model.b["1"].a = Var(bounds=(-100, 100), initialize=2)
        b = model.b["1"].b = Var(bounds=(-100, 100), initialize=20)
        model.b["1"].c = Constraint(expr=b == 10 * a)
        a.fix(2)
        return model

    def setup_model02(self):
        model = ConcreteModel()
        a = model.a = Param(default=1, mutable=True)
        b = model.b = Param(default=2, mutable=True)
        c = model.c = Param(initialize=4)
        e = model.e = Expression(expr=a + b)
        x = model.x = Var([1, 2], initialize={1: 1.5, 2: 2.5}, bounds=(-10, 10))
        model.f = Objective(expr=(x[1] - a) ** 2 + (x[2] - b) ** 2)
        model.g = Constraint(expr=x[1] + x[2] - c >= 0)
        model.suf1 = Suffix()
        model.dual = Suffix(direction=Suffix.IMPORT)
        model.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
        model.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
        return model

    def setup_model02b(self):
        model = ConcreteModel()
        model.p = Param(["a", "b"], default={"a": 1, "b": 2}, mutable=True)
        a = model.p["a"]
        b = model.p["b"]
        c = model.c = Param(initialize=4)
        x = model.x = Var(["1", "2"], initialize={"1": 1.5, "2": 2.5}, bounds=(-10, 10))
        model.f = Objective(expr=(x["1"] - a) ** 2 + (x["2"] - b) ** 2)
        model.g = Constraint(expr=x["1"] + x["2"] - c >= 0)
        model.dual = Suffix(direction=Suffix.IMPORT)
        model.ipopt_zL_out = Suffix(direction=Suffix.IMPORT)
        model.ipopt_zU_out = Suffix(direction=Suffix.IMPORT)
        return model

    def setup_model03(self):
        m = ConcreteModel()
        m.I = Set(initialize=[1, 2])
        m.J = Set(initialize=[(3, 3), (4, 4)])
        m.J.display()

        @m.Block(m.I)
        def b(b, i):
            b.x = Var(b.model().J, bounds=(i, None))

        m.b[1].x.display()
        m.b[2].x.display()
        m.r = Reference(m.b[:].x[3, :])
        m.r.display()
        return m

    @pytest.mark.unit
    def test01_name(self):
        """
        Simple test of load save json with different model names
        """
        model = self.setup_model01("m1")
        model2 = self.setup_model01("m2")

        model2.b[1].a = 0.11
        model2.b[1].b = 0.11
        model2.x = False
        to_json(model, fname=self.fname, human_read=True)
        from_json(model2, fname=self.fname)
        # make sure they are right
        assert pytest.approx(20) == value(model2.b[1].b)
        assert pytest.approx(2) == value(model2.b[1].a)
        assert value(model2.x) == True

    @pytest.mark.unit
    def test01(self):
        """
        Simple test of load save json
        """
        model = self.setup_model01()
        a = model.b[1].a
        b = model.b[1].b
        to_json(model, fname=self.fname, human_read=True)
        # change variable values
        a.value = 0.11
        b.value = 0.11
        a.unfix()
        model.b[1].deactivate()
        b.setlb(2)
        b.setub(4)
        # reload values
        from_json(model, fname=self.fname)
        # make sure they are right
        assert a.fixed
        assert model.b[1].active
        assert value(b) == 20
        assert value(a) == 2
        assert b.lb == -100
        assert b.ub == 100

    @pytest.mark.unit
    def test01b(self):
        """
        Simple test of load save json
        """
        model = self.setup_model01b()
        a = model.b["1"].a
        b = model.b["1"].b
        to_json(model, fname=self.fname, human_read=True)
        # change variable values
        a.value = 0.11
        b.value = 0.11
        a.unfix()
        model.b["1"].deactivate()
        b.setlb(2)
        b.setub(4)
        # reload values
        from_json(model, fname=self.fname)
        # make sure they are right
        assert a.fixed
        assert model.b["1"].active
        assert value(b) == 20
        assert value(a) == 2
        assert b.lb == -100
        assert b.ub == 100

    @pytest.mark.unit
    def test01c(self):
        """
        Simple test of load save json for a _BlockData
        """
        model = self.setup_model01b()
        a = model.b["1"].a
        b = model.b["1"].b
        sd = to_json(model.b["1"], return_dict=True, human_read=True)
        # change variable values
        a.value = 0.11
        b.value = 0.11
        a.unfix()
        model.b["1"].deactivate()
        b.setlb(2)
        b.setub(4)
        # reload values
        from_json(model.b["1"], sd=sd)
        # make sure they are right
        assert a.fixed
        assert model.b["1"].active
        assert value(b) == 20
        assert value(a) == 2
        assert b.lb == -100
        assert b.ub == 100

    @pytest.mark.unit
    def test02(self):
        """Test with suffixes"""
        model = self.setup_model02()
        x = model.x
        model.dual[model.g] = 1
        model.ipopt_zL_out[x[1]] = 0
        model.ipopt_zL_out[x[2]] = 0
        model.ipopt_zU_out[x[1]] = 0
        model.ipopt_zU_out[x[2]] = 0
        model.suf1[model.e] = 222
        to_json(model, fname=self.fname, human_read=True)
        model.x[1].value = 10
        model.x[2].value = 10
        model.suf1[model.e] = 111
        model.dual[model.g] = 10
        model.ipopt_zL_out[x[1]] = 10
        model.ipopt_zL_out[x[2]] = 10
        model.ipopt_zU_out[x[1]] = 10
        model.ipopt_zU_out[x[2]] = 10
        from_json(model, fname=self.fname)
        assert model.suf1[model.e] == 222
        assert value(x[1]) == pytest.approx(1.5)
        assert value(x[2]) == pytest.approx(2.5)
        assert model.dual[model.g] == 1
        assert model.ipopt_zL_out[x[1]] == 0
        assert model.ipopt_zL_out[x[2]] == 0
        assert model.ipopt_zU_out[x[1]] == 0
        assert model.ipopt_zU_out[x[2]] == 0

    @pytest.mark.unit
    def test02b(self):
        """Test with suffixes"""
        model = self.setup_model02b()
        x = model.x
        model.dual[model.g] = 1
        model.ipopt_zL_out[x["1"]] = 0
        model.ipopt_zL_out[x["2"]] = 0
        model.ipopt_zU_out[x["1"]] = 0
        model.ipopt_zU_out[x["2"]] = 0
        to_json(model, fname=self.fname, human_read=True)
        model.x["1"].value = 10
        model.x["2"].value = 10
        model.dual[model.g] = 10
        model.ipopt_zL_out[x["1"]] = 10
        model.ipopt_zL_out[x["2"]] = 10
        model.ipopt_zU_out[x["1"]] = 10
        model.ipopt_zU_out[x["2"]] = 10
        model.p["a"] = 10
        model.p["b"] = 10
        from_json(model, fname=self.fname)
        assert value(model.x["1"]) == pytest.approx(1.5)
        assert value(model.x["2"]) == pytest.approx(2.5)
        assert value(model.p["a"]) == pytest.approx(1)
        assert value(model.p["b"]) == pytest.approx(2)
        assert model.dual[model.g] == pytest.approx(1)
        assert model.ipopt_zL_out[x["1"]] == pytest.approx(0)
        assert model.ipopt_zL_out[x["2"]] == pytest.approx(0)
        assert model.ipopt_zU_out[x["1"]] == pytest.approx(0)
        assert model.ipopt_zU_out[x["2"]] == pytest.approx(0)

    @pytest.mark.unit
    def test02c(self):
        """Test with suffixes"""
        model = self.setup_model02b()
        model.dual[model.g] = 222
        wts = StoreSpec(suffix=False)
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        model.dual[model.g] = 111
        from_json(model, fname=self.fname, wts=wts)
        assert model.dual[model.g] == 111

    @pytest.mark.unit
    def test02d(self):
        """Test with suffixes"""
        model = self.setup_model02b()
        model.dual[model.g] = 222
        wts = StoreSpec(suffix=True)
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        model.dual[model.g] = 111
        from_json(model, fname=self.fname)
        assert model.dual[model.g] == 222

    @pytest.mark.unit
    def test02e(self):
        """Test with suffixes"""
        model = self.setup_model02b()
        model.ipopt_zL_out[model.x["1"]] = 222
        model.dual[model.g] = 10
        wts = StoreSpec(suffix_filter="dual")
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        model.ipopt_zL_out[model.x["1"]] = 111
        model.dual[model.g] = 11
        from_json(model, fname=self.fname)
        assert model.ipopt_zL_out[model.x["1"]] == 111
        assert model.dual[model.g] == 10

    @pytest.mark.unit
    def test03(self):
        """
        This tests a StoreSpec object meant for initialization.  It reloads
        the saved state of variable values and whether they are fixed or
        unfixed but it only loads values for variables that were originally
        fixed. It's use would be in doing initialization that changes which
        variables are fixed and does whatever, but when the original state is
        reloaded only the original values for fixed variables are reloaded.  You
        basically end up with the same problem, just a different initial guess.
        """
        model = self.setup_model02()
        x = model.x
        x[1].fix(1)
        wts = StoreSpec.value_isfixed(only_fixed=True)
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        x[1].unfix()
        x[1].value = 2
        x[2].value = 10
        from_json(model, fname=self.fname, wts=wts)
        assert x[1].fixed
        assert value(x[1]) == 1
        assert value(x[2]) == 10

    @pytest.mark.unit
    def test04(self):
        """
        Like test03, but this StoreSpec also saves/loads active/deactivated
        component attribute and parameter values.
        """
        model = self.setup_model02()
        x = model.x
        x[1].fix(1)
        wts = StoreSpec.value_isfixed_isactive(only_fixed=True)
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        x[1].unfix()
        x[1].value = 2
        x[2].value = 10
        model.g.deactivate()
        from_json(model, fname=self.fname, wts=wts)
        assert x[1].fixed
        assert value(x[1]) == 1
        assert value(x[2]) == 10
        assert model.g.active

    @pytest.mark.unit
    def test04b(self):
        # test old style StoreSpec args
        wts = StoreSpec(
            classes=(
                (Var, (), None),
                (BooleanVar, (), None),
                (Param, (), None),
                (Constraint, ("active",), None),
                (Block, ("active",), None),
            ),
            data_classes=(
                (Var._ComponentDataClass, ("value", "fixed"), _only_fixed),
                (BooleanVar._ComponentDataClass, ("value", "fixed"), _only_fixed),
                (pyomo.core.base.param.ParamData, ("value",), None),
                (Constraint._ComponentDataClass, ("active",), None),
                (Block._ComponentDataClass, ("active",), None),
            ),
        )
        model = self.setup_model02()
        x = model.x
        x[1].fix(1)
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        x[1].unfix()
        x[1].value = 2
        x[2].value = 10
        model.g.deactivate()
        from_json(model, fname=self.fname, wts=wts)
        assert x[1].fixed
        assert value(x[1]) == 1
        assert value(x[2]) == 10
        assert model.g.active

    @pytest.mark.unit
    def test05(self):
        """Try just saving values"""
        model = self.setup_model02()
        model.x[1].value = 1
        wts = StoreSpec.value()
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        model.g.deactivate()
        model.x[1].setlb(-4)
        model.x[1].value = 3
        model.x[2].value = 6
        from_json(model, fname=self.fname, wts=wts)
        assert value(model.x[1]) == 1
        assert model.x[1].lb == -4
        assert value(model.x[2]) == pytest.approx(2.5)
        assert not model.g.active

    @pytest.mark.unit
    def test05b(self):
        """Try just saving values with fixed filter"""
        model = self.setup_model02()
        model.x[1].fix(1)
        wts = StoreSpec.value(only_not_fixed=True)
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        model.g.deactivate()
        model.x[1].setlb(-4)
        model.x[1].value = 3
        model.x[2].value = 6
        from_json(model, fname=self.fname, wts=wts)
        assert value(model.x[1]) == 3  # should not load since is fixed
        assert model.x[1].lb == -4
        assert value(model.x[2]) == pytest.approx(2.5)
        assert not model.g.active

    @pytest.mark.unit
    def test06(self):
        """Try just saving bounds"""
        model = self.setup_model02()
        model.x[1].value = 1
        wts = StoreSpec.bound()
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        model.g.deactivate()
        model.x[1].setlb(-4)
        model.x[1].value = 3
        model.x[2].value = 6
        from_json(model, fname=self.fname, wts=wts)
        assert value(model.x[1]) == 3
        assert model.x[1].lb == -10
        assert value(model.x[2]) == 6
        assert not model.g.active

    @pytest.mark.unit
    def test07(self):
        """Try just saving just if fixed"""
        model = self.setup_model02()
        model.x[1].fix(1)
        wts = StoreSpec.isfixed()
        to_json(model, fname=self.fname, human_read=True, wts=wts)
        model.g.deactivate()
        model.x[1].setlb(-4)
        model.x[1].unfix()
        model.x[2].fix(6)
        from_json(model, fname=self.fname, wts=wts)
        assert value(model.x[1]) == 1
        assert model.x[1].lb == -4
        assert value(model.x[2]) == 6
        assert model.x[1].fixed
        assert not model.x[2].fixed
        assert not model.g.active

    @pytest.mark.unit
    def test08(self):
        """Try just saving suffixes"""
        model = self.setup_model02()
        model.x[1].fix(1)

        model.dual[model.g] = 1
        model.ipopt_zL_out[model.x[1]] = 1
        model.ipopt_zL_out[model.x[2]] = 1
        model.ipopt_zU_out[model.x[1]] = 1
        model.ipopt_zU_out[model.x[2]] = 1

        to_json(model, fname=self.fname, wts=StoreSpec.suffix())

        model.dual[model.g] = 10
        model.ipopt_zL_out[model.x[1]] = 10
        model.ipopt_zL_out[model.x[2]] = 10
        model.ipopt_zU_out[model.x[1]] = 10
        model.ipopt_zU_out[model.x[2]] = 10
        model.g.deactivate()
        model.x[1].setlb(-4)
        model.x[1].unfix()
        model.x[2].fix(6)

        from_json(model, fname=self.fname, wts=StoreSpec.suffix())
        assert value(model.x[1]) == 1
        assert value(model.x[2]) == 6
        assert not model.x[1].fixed
        assert model.x[2].fixed
        assert not model.g.active
        assert model.dual[model.g] == 1
        assert model.ipopt_zL_out[model.x[1]] == 1
        assert model.ipopt_zL_out[model.x[2]] == 1
        assert model.ipopt_zU_out[model.x[1]] == 1
        assert model.ipopt_zU_out[model.x[2]] == 1
        assert model.x[1].lb == -4

    @pytest.mark.unit
    def test09(self):
        """Try just saving suffixes, and suffix filter"""
        model = self.setup_model02()

        model.dual[model.g] = 1
        model.ipopt_zL_out[model.x[1]] = 1
        model.ipopt_zL_out[model.x[2]] = 1
        model.ipopt_zU_out[model.x[1]] = 1
        model.ipopt_zU_out[model.x[2]] = 1

        wts = StoreSpec.suffix(suffix_filter=("dual",))
        to_json(model, fname=self.fname, wts=wts)

        model.dual[model.g] = 10
        model.ipopt_zL_out[model.x[1]] = 10
        model.ipopt_zL_out[model.x[2]] = 10
        model.ipopt_zU_out[model.x[1]] = 10
        model.ipopt_zU_out[model.x[2]] = 10

        from_json(model, fname=self.fname, wts=wts)
        assert model.dual[model.g] == 1
        assert model.ipopt_zL_out[model.x[1]] == 10
        assert model.ipopt_zL_out[model.x[2]] == 10
        assert model.ipopt_zU_out[model.x[1]] == 10
        assert model.ipopt_zU_out[model.x[2]] == 10

    @pytest.mark.unit
    def test10(self):
        """Try just saving suffixes, and suffix filter only on write"""
        model = self.setup_model02()

        model.dual[model.g] = 1
        model.ipopt_zL_out[model.x[1]] = 1
        model.ipopt_zL_out[model.x[2]] = 1
        model.ipopt_zU_out[model.x[1]] = 1
        model.ipopt_zU_out[model.x[2]] = 1

        wts = StoreSpec.suffix(suffix_filter=("dual",))
        to_json(model, fname=self.fname, wts=wts)

        model.dual[model.g] = 10
        model.ipopt_zL_out[model.x[1]] = 10
        model.ipopt_zL_out[model.x[2]] = 10
        model.ipopt_zU_out[model.x[1]] = 10
        model.ipopt_zU_out[model.x[2]] = 10

        from_json(model, fname=self.fname, wts=StoreSpec.suffix())
        assert model.dual[model.g] == 1
        assert model.ipopt_zL_out[model.x[1]] == 10
        assert model.ipopt_zL_out[model.x[2]] == 10
        assert model.ipopt_zU_out[model.x[1]] == 10
        assert model.ipopt_zU_out[model.x[2]] == 10

    @pytest.mark.unit
    def test11(self):
        """Try just saving suffixes, and suffix filter only on read"""
        model = self.setup_model02()

        model.dual[model.g] = 1
        model.ipopt_zL_out[model.x[1]] = 1
        model.ipopt_zL_out[model.x[2]] = 1
        model.ipopt_zU_out[model.x[1]] = 1
        model.ipopt_zU_out[model.x[2]] = 1

        to_json(model, fname=self.fname, wts=StoreSpec.suffix())

        model.dual[model.g] = 10
        model.ipopt_zL_out[model.x[1]] = 10
        model.ipopt_zL_out[model.x[2]] = 10
        model.ipopt_zU_out[model.x[1]] = 10
        model.ipopt_zU_out[model.x[2]] = 10

        wts = StoreSpec.suffix(suffix_filter=("dual",))
        from_json(model, fname=self.fname, wts=wts)
        assert model.dual[model.g] == 1
        assert model.ipopt_zL_out[model.x[1]] == 10
        assert model.ipopt_zL_out[model.x[2]] == 10
        assert model.ipopt_zU_out[model.x[1]] == 10
        assert model.ipopt_zU_out[model.x[2]] == 10

    @pytest.mark.unit
    def test12(self):
        """Test some odd set compoenents"""
        model = self.setup_model03()
        self.assertIs(model.r.ctype, Var)
        model.r[1, 3] = 1
        model.r[2, 3] = 3
        d = to_json(model, return_dict=True, human_read=True)
        model.r[1, 3] = 6
        model.r[2, 3] = 8
        assert value(model.b[1].x[3, 3]) == 6
        assert value(model.b[2].x[3, 3]) == 8
        from_json(model, sd=d)
        assert value(model.b[1].x[3, 3]) == 1
        assert value(model.b[2].x[3, 3]) == 3


if __name__ == "__main__":
    unittest.main()



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
This module contains tests for the variable replace transformation.
"""

import pytest
import pyomo.environ as pyo
import idaes.core.plugins
from idaes.core.util.model_statistics import degrees_of_freedom
from pyomo.repn import generate_standard_repn

__author__ = "John Eslick"


@pytest.fixture
def model():
    m = pyo.ConcreteModel()
    m.x = pyo.Var([1, 2, 3, 4], initialize=2)
    m.y = pyo.Var(initialize=3)
    m.z = pyo.Var([1, 2, 3, 4], initialize=0)
    m.e = pyo.Expression(expr=m.x[1] + m.x[2])
    m.z.fix()
    m.z[4] = 4.0

    # In the first round this becomes x[1].fix(m.z[1])
    m.c1 = pyo.Constraint(expr=m.z[1] == m.x[1])
    # In the first round this beconmes y.fix(m.z[2])
    m.c2 = pyo.Constraint(expr=m.z[2] == m.y)
    # In the second round this becomes x[2].fix(m.z[3]-m.x[1])
    m.c3 = pyo.Constraint(expr=m.z[3] == m.e)
    # This constraint is nonlinear but m.x[1] will be fixed
    m.c4 = pyo.Constraint(expr=m.z[4] == m.x[3] ** 2 + m.x[1])
    # This constraint is eliminated by a substitution, x[3] and x[4] are not
    # fixed, so it's a linear equation with two variables.  This should be in the
    # second round after x[1] becomes fixed in the first
    m.c5 = pyo.Constraint(expr=m.z[4] == m.x[4] + m.x[1] + m.x[3])

    # Solution
    m.x[1] = 0
    m.x[2] = 0
    m.x[3] = 2
    m.x[4] = 2
    m.y = 0

    # make a dict to check variable values against (vars and named expressions)
    m.sol = {}
    for v in m.component_data_objects((pyo.Var, pyo.Expression)):
        m.sol[id(v)] = pyo.value(v)
    return m


@pytest.mark.unit
def test_model(model):
    m = model
    assert degrees_of_freedom(m) == 0
    assert len([c for c in m.component_data_objects(pyo.Constraint, active=True)]) == 5


@pytest.mark.unit
def test_transform(model):
    m = model
    elim = pyo.TransformationFactory("simple_equality_eliminator")
    elim.apply_to(m)
    assert degrees_of_freedom(m) == 0

    assert len([c for c in m.component_data_objects(pyo.Constraint, active=True)]) == 1
    assert m.x[1].fixed
    assert m.x[2].fixed
    assert m.y.fixed


@pytest.mark.unit
def test_revert_constraint():
    # test that multiple replaced constraints are reverted right
    m = pyo.ConcreteModel()
    m.x = pyo.Var([1, 2, 3, 4], initialize={1: 0, 2: 2, 3: 3, 4: 4})
    m.x[1].unfix()
    # This constraint should be replaced twice
    # 1.) x[2] and x[3] get written in terms of x[4]
    # 2.) x[1] == 0.5*x[4] + x[3]
    # 3.) x[1] == 0.5*x[4] + 0.75*x[4] (1.25*x4)
    m.c1 = pyo.Constraint(expr=m.x[1] - m.x[2] - m.x[3] == 0)
    m.c2 = pyo.Constraint(expr=m.x[2] * 2 - m.x[4] == 0)
    m.c3 = pyo.Constraint(expr=m.x[3] * 4 - m.x[4] * 3 == 0)
    orig = {id(m.x[1]): 1, id(m.x[2]): -1, id(m.x[3]): -1}
    tf = {id(m.x[1]): 1, id(m.x[4]): -1.25}

    elim = pyo.TransformationFactory("simple_equality_eliminator")
    elim.apply_to(m)
    # Check that numbers indicate correctly transformed constraint
    assert pytest.approx(pyo.value(m.c1.body - m.c1.lower)) == -5
    # Check the linear representation of the constraint to make sure it's right
    c1_repn = generate_standard_repn(m.c1.body)
    for i, v in enumerate(c1_repn.linear_vars):
        assert id(v) in tf
        assert c1_repn.linear_coefs[i] == tf[id(v)]
    assert len(c1_repn.linear_vars) == 2

    elim.revert()
    # make sure the constraint is back to m.x[1] == m.x[2] + m.x[3]
    m.x[1] = 0.0
    m.x[2] = 0.5
    m.x[3] = 0.75
    assert pytest.approx(pyo.value(m.c1.body - m.c1.lower)) == -1.25
    c1_repn = generate_standard_repn(m.c1.body)
    for i, v in enumerate(c1_repn.linear_vars):
        assert id(v) in orig
        assert c1_repn.linear_coefs[i] == orig[id(v)]
    assert len(c1_repn.linear_vars) == 3

    # check again with fixed var
    m.x[4].fix()
    elim.apply_to(m, max_iter=3)
    assert len([c for c in m.component_data_objects(pyo.Constraint, active=True)]) == 0
    # Since x[4] is fixed all the constraints can be eliminated and all the variables
    # can be fixed based on the value of x[4].  First make sure x[1] is fixed right
    assert m.x[1].fixed
    assert pytest.approx(pyo.value(m.x[1])) == 5
    elim.revert()
    # Now when we revert the transformation, make sure the constraints come back
    # correctly and that the variables are not fixed anymore.
    m.x[1] = 0.0
    m.x[2] = 0.5
    m.x[3] = 0.75
    assert len([c for c in m.component_data_objects(pyo.Constraint, active=True)]) == 3
    assert pytest.approx(pyo.value(m.c1.body - m.c1.lower)) == -1.25
    assert not m.x[1].fixed
    c1_repn = generate_standard_repn(m.c1.body)
    for i, v in enumerate(c1_repn.linear_vars):
        assert id(v) in orig
        assert c1_repn.linear_coefs[i] == orig[id(v)]
    assert len(c1_repn.linear_vars) == 3


@pytest.mark.unit
def test_revert_constraint_with_named_expr():
    # test that named Expressions get reverted right.
    m = pyo.ConcreteModel()
    m.x = pyo.Var([1, 2, 3, 4], initialize={1: 0, 2: 2, 3: 3, 4: 4})
    m.x[1].unfix()
    m.e1 = pyo.Expression(expr=m.x[2])
    m.c1 = pyo.Constraint(expr=m.x[1] - m.e1 - m.x[3] == 0)
    m.c2 = pyo.Constraint(expr=m.x[2] - m.x[4] / 2 == 0)
    m.c3 = pyo.Constraint(expr=m.x[3] * 4 - m.x[4] * 3 == 0)
    elim = pyo.TransformationFactory("simple_equality_eliminator")
    elim.apply_to(m, max_iter=3)
    assert pytest.approx(pyo.value(m.c1.body - m.c1.lower)) == -5
    assert pytest.approx(pyo.value(m.e1)) == 2  # check that expression gets changed
    elim.revert()
    m.x[1] = 0.0
    m.x[2] = 0.5
    m.x[3] = 0.75
    assert pytest.approx(pyo.value(m.e1)) == 0.5  # check that expression gets reverted
    assert pytest.approx(pyo.value(m.c1.body - m.c1.lower)) == -1.25


@pytest.mark.unit
def test_reverse_var(model):
    m = model
    elim = pyo.TransformationFactory("simple_equality_eliminator")
    elim.apply_to(m)

    # mess up the values for all the substituted vars
    for i, v in elim._subs_map.items():
        v = 1e12
    elim.revert()

    for v in m.component_data_objects((pyo.Var, pyo.Expression)):
        assert pyo.value(v) == m.sol[id(v)]

    assert len([c for c in m.component_data_objects(pyo.Constraint, active=True)]) == 5
    assert not m.x[1].fixed
    assert not m.x[2].fixed
    assert not m.y.fixed


@pytest.mark.unit
def test_bounds(model):
    m = model

    elim = pyo.TransformationFactory("simple_equality_eliminator")

    m.x[3].setlb(-1)
    m.x[3].setub(10)
    m.x[4].setlb(1)
    m.x[4].setub(12)

    elim.apply_to(m)

    # The bounds here come from constraint c5.  z4 is 4, x1 is 0, and x4 and x5
    # are not fixed, so you get 4 = x3 + x4.  I'm not sure if x3 or x4 will be
    # replaced, so I calculated bounds on both.

    # For x3 the bounds that come from x4 are 4 - x4.lb and 4 - x4.ub or (-8 and 3)
    # so the tightest set of bounds from x3 and x4 are (-1, 3)

    # If instead x3 is replaced, the bounds on x4 that come from x3 are
    # 4 - 10 and 4 + 1 or (-6, 5).  The tightest set of bounds on x4 is (1, 5)
    if id(m.x[3]) not in elim._subs_map:
        assert m.x[3].lb == -1
        assert m.x[3].ub == 3
    else:
        assert m.x[4].lb == 1
        assert m.x[4].ub == 5



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tray column model for distillation.
"""

__author__ = "Jaffer Ghouse"

# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
from pyomo.network import Arc, Port
from pyomo.environ import (
    Integers,
    RangeSet,
    TransformationFactory,
    Block,
    Reference,
    Var,
    Constraint,
)

# Import IDAES cores
from idaes.models_extra.column_models import Tray, Condenser, Reboiler
from idaes.models_extra.column_models.condenser import CondenserType, TemperatureSpec
from idaes.core import declare_process_block_class, UnitModelBlockData, useDefault
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.config import is_physical_parameter_block
from idaes.core.solvers import get_solver

import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog
from idaes.core.util.initialization import propagate_state

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("TrayColumn")
class TrayColumnData(UnitModelBlockData):
    """
    Tray Column model.
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
**default** = False. Tray column units do not support dynamic behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
**default** - False. Tray column units do not have defined volume, thus
this must be False.""",
        ),
    )
    CONFIG.declare(
        "number_of_trays",
        ConfigValue(
            default=None,
            domain=In(Integers),
            description="Number of trays in the column",
            doc="""Indicates the number of trays to be constructed.
**default** - None.
**Valid values:** {
Must be integer number.}""",
        ),
    )
    CONFIG.declare(
        "feed_tray_location",
        ConfigValue(
            default=None,
            domain=In(Integers),
            description="Feed tray location in the column",
            doc="""Indicates the number of trays to be constructed.
**default** - None.
**Valid values:** {
Must be integer number.}""",
        ),
    )
    CONFIG.declare(
        "condenser_type",
        ConfigValue(
            default=CondenserType.totalCondenser,
            domain=In(CondenserType),
            description="Type of condenser flag",
            doc="""Indicates what type of condenser should be constructed,
**default** - CondenserType.totalCondenser.
**Valid values:** {
**CondenserType.totalCondenser** - Incoming vapor from top tray is condensed
to all liquid,
**CondenserType.partialCondenser** - Incoming vapor from top tray is
partially condensed to a vapor and liquid stream.}""",
        ),
    )
    CONFIG.declare(
        "condenser_temperature_spec",
        ConfigValue(
            default=None,
            domain=In(TemperatureSpec),
            description="Temperature spec for the condenser",
            doc="""Temperature specification for the condenser,
**default** - TemperatureSpec.none
**Valid values:** {
**TemperatureSpec.none** - No spec is selected,
**TemperatureSpec.atBubblePoint** - Condenser temperature set at
bubble point i.e. total condenser,
**TemperatureSpec.customTemperature** - Condenser temperature at
user specified temperature.}""",
        ),
    )
    CONFIG.declare(
        "has_heat_transfer",
        ConfigValue(
            default=False,
            domain=Bool,
            description="heat duty to/from tray construction flag.",
            doc="""indicates if there is heat duty to/from the tray,
**default** - False.
**Valid values:** {
**True** - include a heat duty term,
**False** - exclude a heat duty term.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=Bool,
            description="pressure change term construction flag",
            doc="""indicates whether terms for pressure change should be
constructed,
**default** - False.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "has_liquid_side_draw",
        ConfigValue(
            default=False,
            domain=Bool,
            description="liquid side draw construction flag.",
            doc="""indicates if there is a liquid side draw from all trays,
**default** - False.
**Valid values:** {
**True** - include a liquid side draw for all trays,
**False** - exclude a liquid side draw for all trays.}""",
        ),
    )
    CONFIG.declare(
        "has_vapor_side_draw",
        ConfigValue(
            default=False,
            domain=Bool,
            description="vapor side draw construction flag.",
            doc="""indicates if there is a vapor side draw from all trays,
**default** - False.
**Valid values:** {
**True** - include a vapor side draw for all trays,
**False** - exclude a vapor side draw for all trays.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="property package to use for control volume",
            doc="""property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="arguments to use for constructing property packages",
            doc="""a ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """Build the model.
        Args:
            None
        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(TrayColumnData, self).build()

        # Create set for constructing indexed trays
        if self.config.number_of_trays is not None:
            self.tray_index = RangeSet(1, self.config.number_of_trays)
            self._rectification_index = RangeSet(1, self.config.feed_tray_location - 1)
            self._stripping_index = RangeSet(
                self.config.feed_tray_location + 1, self.config.number_of_trays
            )

        else:
            raise ConfigurationError(
                "The config argument number_of_trays "
                "needs to specified and cannot be None."
            )

        # Add trays
        # TODO:
        # 1. Add support for multiple feed tray locations
        # 2. Add support for specifying which trays have side draws
        self.rectification_section = Tray(
            self._rectification_index,
            has_liquid_side_draw=self.config.has_liquid_side_draw,
            has_vapor_side_draw=self.config.has_vapor_side_draw,
            has_heat_transfer=self.config.has_heat_transfer,
            has_pressure_change=self.config.has_pressure_change,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.feed_tray = Tray(
            is_feed_tray=True,
            has_liquid_side_draw=self.config.has_liquid_side_draw,
            has_vapor_side_draw=self.config.has_vapor_side_draw,
            has_heat_transfer=self.config.has_heat_transfer,
            has_pressure_change=self.config.has_pressure_change,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.stripping_section = Tray(
            self._stripping_index,
            has_liquid_side_draw=self.config.has_liquid_side_draw,
            has_vapor_side_draw=self.config.has_vapor_side_draw,
            has_heat_transfer=self.config.has_heat_transfer,
            has_pressure_change=self.config.has_pressure_change,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        # Add condenser
        self.condenser = Condenser(
            condenser_type=self.config.condenser_type,
            temperature_spec=self.config.condenser_temperature_spec,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        # Add reboiler
        self.reboiler = Reboiler(
            has_boilup_ratio=True,
            has_pressure_change=self.config.has_pressure_change,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        # Add extension to the feed port
        self.feed = Port(extends=self.feed_tray.feed)

        # Add extensions to rectification section

        # Add extensions to stripping section

        # Construct arcs between trays, condenser, and reboiler
        self._make_rectification_arcs()
        self._make_stripping_arcs()
        self._make_feed_arcs()
        self._make_condenser_arcs()
        self._make_reboiler_arcs()

        TransformationFactory("network.expand_arcs").apply_to(self)

        for var in self.component_data_objects(Var, descend_into=True):
            if "pressure" in var.name:
                iscale.set_scaling_factor(var, 1e-5)
            if "temperature" in var.name:
                iscale.set_scaling_factor(var, 1e-2)
            if "enth_mol_phase" in var.name:
                iscale.set_scaling_factor(var, 1e-4)
            if "heat" in var.name:
                iscale.set_scaling_factor(var, 1e-5)
            if "flow_mol_phase_comp" in var.name:
                iscale.set_scaling_factor(var, 1e1)
            if "_t1" in var.name:
                iscale.set_scaling_factor(var, 1e-2)

        for constr in self.component_data_objects(Constraint, descend_into=True):
            if "pressure_drop_equation" in constr.name:
                iscale.constraint_scaling_transform(constr, 1e-2, overwrite=False)
            if "eq_P_vap" in constr.name:
                iscale.constraint_scaling_transform(constr, 1e-2, overwrite=False)

    def _make_rectification_arcs(self):
        self._rectification_stream_index = RangeSet(
            1, self.config.feed_tray_location - 2
        )

        def rule_liq_stream(self, i):
            return {
                "source": self.rectification_section[i].liq_out,
                "destination": self.rectification_section[i + 1].liq_in,
            }

        def rule_vap_stream(self, i):
            return {
                "source": self.rectification_section[i + 1].vap_out,
                "destination": self.rectification_section[i].vap_in,
            }

        self.rectification_liq_stream = Arc(
            self._rectification_stream_index, rule=rule_liq_stream
        )
        self.rectification_vap_stream = Arc(
            self._rectification_stream_index, rule=rule_vap_stream
        )

    def _make_stripping_arcs(self):

        self._stripping_stream_index = RangeSet(
            self.config.feed_tray_location + 1, self.config.number_of_trays - 1
        )

        def rule_liq_stream(self, i):
            return {
                "source": self.stripping_section[i].liq_out,
                "destination": self.stripping_section[i + 1].liq_in,
            }

        def rule_vap_stream(self, i):
            return {
                "source": self.stripping_section[i + 1].vap_out,
                "destination": self.stripping_section[i].vap_in,
            }

        self.stripping_liq_stream = Arc(
            self._stripping_stream_index, rule=rule_liq_stream
        )
        self.stripping_vap_stream = Arc(
            self._stripping_stream_index, rule=rule_vap_stream
        )

    def _make_feed_arcs(self):

        self.feed_liq_in = Arc(
            source=self.rectification_section[
                self.config.feed_tray_location - 1
            ].liq_out,
            destination=self.feed_tray.liq_in,
        )

        self.feed_liq_out = Arc(
            source=self.feed_tray.liq_out,
            destination=self.stripping_section[
                self.config.feed_tray_location + 1
            ].liq_in,
        )

        self.feed_vap_in = Arc(
            source=self.stripping_section[self.config.feed_tray_location + 1].vap_out,
            destination=self.feed_tray.vap_in,
        )

        self.feed_vap_out = Arc(
            source=self.feed_tray.vap_out,
            destination=self.rectification_section[
                self.config.feed_tray_location - 1
            ].vap_in,
        )

    def _make_condenser_arcs(self):

        self.condenser_vap_in = Arc(
            source=self.rectification_section[1].vap_out,
            destination=self.condenser.inlet,
        )

        self.condenser_reflux_out = Arc(
            source=self.condenser.reflux,
            destination=self.rectification_section[1].liq_in,
        )

    def _make_reboiler_arcs(self):

        self.reboiler_liq_in = Arc(
            source=self.stripping_section[self.config.number_of_trays].liq_out,
            destination=self.reboiler.inlet,
        )

        self.reboiler_vap_out = Arc(
            source=self.reboiler.vapor_reboil,
            destination=self.stripping_section[self.config.number_of_trays].vap_in,
        )

    def initialize(
        self,
        state_args_feed=None,
        state_args_liq=None,
        state_args_vap=None,
        solver=None,
        optarg=None,
        outlvl=idaeslog.NOTSET,
    ):

        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        init_log.info("Begin initialization.")

        solverobj = get_solver(solver, optarg)

        feed_flags = self.feed_tray.initialize(
            solver=solver, optarg=optarg, outlvl=outlvl
        )

        propagate_state(destination=self.condenser.inlet, source=self.feed_tray.vap_out)

        self.condenser.initialize(solver=solver, optarg=optarg, outlvl=outlvl)

        propagate_state(destination=self.reboiler.inlet, source=self.feed_tray.liq_out)

        self.reboiler.initialize(solver=solver, optarg=optarg, outlvl=outlvl)

        # initialize the rectification section
        for i in self._rectification_index:
            propagate_state(
                destination=self.rectification_section[i].liq_in,
                source=self.condenser.reflux,
            )
            propagate_state(
                destination=self.rectification_section[i].vap_in,
                source=self.feed_tray.vap_out,
            )
            if i == 1:
                rect_liq_flags = self.rectification_section[i].initialize(
                    hold_state_liq=True,
                    hold_state_vap=False,
                    solver=solver,
                    optarg=optarg,
                    outlvl=outlvl,
                )
            elif i == len(self._rectification_index):
                rect_vap_flags = self.rectification_section[i].initialize(
                    hold_state_liq=False,
                    hold_state_vap=True,
                    solver=solver,
                    optarg=optarg,
                    outlvl=outlvl,
                )
            else:
                self.rectification_section[i].initialize(
                    solver=solver, optarg=optarg, outlvl=outlvl
                )

        # initialize the stripping section
        for i in self._stripping_index:
            propagate_state(
                destination=self.stripping_section[i].liq_in,
                source=self.feed_tray.liq_out,
            )
            propagate_state(
                destination=self.stripping_section[i].vap_in,
                source=self.reboiler.vapor_reboil,
            )
            if i == self.config.feed_tray_location + 1:
                strip_liq_flags = self.stripping_section[i].initialize(
                    hold_state_liq=True,
                    hold_state_vap=False,
                    solver=solver,
                    optarg=optarg,
                    outlvl=outlvl,
                )
            elif i == self.config.number_of_trays:
                strip_vap_flags = self.stripping_section[i].initialize(
                    hold_state_liq=False,
                    hold_state_vap=True,
                    solver=solver,
                    optarg=optarg,
                    outlvl=outlvl,
                )
            else:
                self.stripping_section[i].initialize(
                    solver=None, optarg=optarg, outlvl=outlvl
                )

        # For initialization purposes and to enable solving individual sections
        # creating a temp block. Note that this temp block is a reference to
        # the rectification, stripping, and feed sections. Also, expanded arcs
        # are added to the temp block as the initialization solve proceeds.
        self._temp_block = Block()

        self._temp_block.rectification = Block()

        # adding reference to the rectification section and the expanded
        # vapor and liquid arcs
        self._temp_block.rectification.trays = Reference(self.rectification_section)
        self._temp_block.rectification.expanded_liq_stream = Reference(
            self.rectification_liq_stream[:].expanded_block
        )
        self._temp_block.rectification.expanded_vap_stream = Reference(
            self.rectification_vap_stream[:].expanded_block
        )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self._temp_block.rectification, tee=slc.tee)
        init_log.info(
            "Rectification section initialization status {}.".format(
                idaeslog.condition(res)
            )
        )

        self._temp_block.stripping = Block()

        # adding reference to the stripping section and the expanded
        # vapor and liquid arcs
        self._temp_block.stripping.trays = Reference(self.stripping_section)
        self._temp_block.stripping.expanded_liq_stream = Reference(
            self.stripping_liq_stream[:].expanded_block
        )
        self._temp_block.stripping.expanded_vap_stream = Reference(
            self.stripping_vap_stream[:].expanded_block
        )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self._temp_block.stripping, tee=slc.tee)
        init_log.info(
            "Stripping section initialization status {}.".format(
                idaeslog.condition(res)
            )
        )

        # releasing the fixed inlets for the vap in to the rectification
        # to enable connection with the feed tray vap out
        self.rectification_section[
            len(self._rectification_index)
        ].properties_in_vap.release_state(flags=rect_vap_flags, outlvl=outlvl)

        # releasing the fixed inlets for the liq in to the stripping
        # to enable connection with the feed tray liq out
        self.stripping_section[
            self.config.feed_tray_location + 1
        ].properties_in_liq.release_state(flags=strip_liq_flags, outlvl=outlvl)

        # Adding the feed tray to temp block solve
        self._temp_block.feed_tray = Reference(self.feed_tray)
        self._temp_block.expanded_feed_liq_stream_in = Reference(
            self.feed_liq_in.expanded_block
        )
        self._temp_block.expanded_feed_liq_stream_out = Reference(
            self.feed_liq_out.expanded_block
        )
        self._temp_block.expanded_feed_vap_stream_in = Reference(
            self.feed_vap_in.expanded_block
        )
        self._temp_block.expanded_feed_vap_stream_out = Reference(
            self.feed_vap_out.expanded_block
        )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self._temp_block, tee=slc.tee)
        init_log.info(
            "Column section initialization status {}.".format(idaeslog.condition(res))
        )

        self.rectification_section[1].properties_in_liq.release_state(
            flags=rect_liq_flags, outlvl=outlvl
        )

        # Adding the condenser to the temp block solve
        self._temp_block.condenser = Reference(self.condenser)
        self._temp_block.expanded_condenser_vap_in = Reference(
            self.condenser_vap_in.expanded_block
        )
        self._temp_block.expanded_condenser_reflux_out = Reference(
            self.condenser_reflux_out.expanded_block
        )

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self._temp_block, tee=slc.tee)
        init_log.info(
            "Column section + Condenser initialization status {}.".format(
                idaeslog.condition(res)
            )
        )

        self.stripping_section[
            self.config.number_of_trays
        ].properties_in_vap.release_state(flags=strip_vap_flags, outlvl=outlvl)

        # Delete the _temp_block as next solve is solving the entire column.
        # If we add the reboiler to the temp block, it will be similar to
        # solving the original block. This ensures that if
        # initialize is triggered twice, there is no implicit replacing
        # component error.
        self.del_component(self._temp_block)

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = solverobj.solve(self, tee=slc.tee)
        init_log.info(
            "Column section + Condenser + Reboiler initialization status {}.".format(
                idaeslog.condition(res)
            )
        )

        # release feed tray state once initialization is complete
        self.feed_tray.properties_in_feed.release_state(flags=feed_flags, outlvl=outlvl)

        # TODO : This fails in the current model
        # if not check_optimal_termination(res):
        #     raise InitializationError(
        #         f"{self.name} failed to initialize successfully. Please check "
        #         f"the output logs for more information.")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Benzene-Toluene phase equilibrium package for liquid and vapor using the
Peng-Robinson equation of state.

Example property package using the Generic Property Package Framework.
This exmample shows how to set up a property package to do benzene-toluene
phase equilibrium in the generic framework using the Peng-Robinson equation of
state for liquid and vapor phases along with methods drawn from the pre-built
IDAES property libraries.
"""
# Import Python libraries
import logging

# Import Pyomo units
from pyomo.environ import units as pyunits

# Import IDAES cores
from idaes.core import LiquidPhase, VaporPhase, Component

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.phase_equil import (
    CubicComplementarityVLE,
)
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    LogBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import log_fugacity
from idaes.models.properties.modular_properties.pure import RPP4


# Set up logger
_log = logging.getLogger(__name__)


# ---------------------------------------------------------------------
# Configuration dictionary for an ideal Benzene-Toluene system

# Data Sources:
# [1] The Properties of Gases and Liquids (1987)
#     4th edition, Chemical Engineering Series - Robert C. Reid
# [3] Engineering Toolbox, https://www.engineeringtoolbox.com
#     Retrieved 1st December, 2019

configuration = {
    # Specifying components
    "components": {
        "benzene": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (78.1136e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (48.9e5, pyunits.Pa),  # [1]
                "temperature_crit": (562.2, pyunits.K),  # [1]
                "omega": 0.212,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (-3.392e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (4.739e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-3.017e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (7.130e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (82.9e3, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    -269,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (-6.98273, None),  # [1]
                    "B": (1.33213, None),
                    "C": (-2.62863, None),
                    "D": (-3.33399, None),
                },
            },
        },
        "toluene": {
            "type": Component,
            "enth_mol_ig_comp": RPP4,
            "entr_mol_ig_comp": RPP4,
            "pressure_sat_comp": RPP4,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (92.1405e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (41e5, pyunits.Pa),  # [1]
                "temperature_crit": (591.8, pyunits.K),  # [1]
                "omega": 0.263,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (-2.435e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (5.125e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-2.765e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (4.911e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (50.1e3, pyunits.J / pyunits.mol),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    -321,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (-7.28607, None),  # [1]
                    "B": (1.38091, None),
                    "C": (-2.83433, None),
                    "D": (-2.79168, None),
                },
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {
            "type": LiquidPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
        "Vap": {
            "type": VaporPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (273.15, 300, 500, pyunits.K),
        "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
    },
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),
    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): CubicComplementarityVLE},
    "bubble_dew_method": LogBubbleDew,
    "parameter_data": {
        "PR_kappa": {
            ("benzene", "benzene"): 0.000,
            ("benzene", "toluene"): 0.000,
            ("toluene", "benzene"): 0.000,
            ("toluene", "toluene"): 0.000,
        }
    },
}
